# Major number means incompatible API
# Minor number means new add-ons but backward compatibility
# Patch number means bugfixes or other improvements that do not affect API
#   with development stage either alpha: a1,a2,a3,.. or beta: b1,b2,b3
__version__ = '1.2'
